import express from 'express'
import swaggerUi from 'swagger-ui-express';
import YAML from 'yamljs';
import { accommodations, reservations, newsletter } from './resources'

import dotenv from "dotenv";
import accommodationRouter from "./routing/accommodations";
import { PrismaClient } from "@prisma/client";
import cookieParser from "cookie-parser";
import checkJwt from "./auth/check-jwt";
import handleBasicAuth from './auth/handle-basic-auth';
import sendMessage from './resources/newsletter';

dotenv.config();

// GET http://localhost:3000/login/token
//    Basic auth - Username: 6b143cf5-dc36-431e-a7a1-202444e75786, Password: << can be empty >>
//    JWT Token from response:
//      eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VySWQiOiI2YjE0M2NmNS1kYzM2LTQzMWUtYTdhMS0yMDI0NDRlNzU3ODYiLCJpYXQiOjE2NTExODgzMjEsImV4cCI6MTY1MTI3NDcyMX0.JhRp7LHD3n0b-fultOFmB-ZyboFR0fjA3x-hGcUHkac
//
// other /api/ endpoints require the JWT token (expect for GET /api/accommodations)
//
// the user we create JWT token for can e.g. use:
// PATCH http://localhost:3000/api/reservations/4b52822e-37cc-498b-b516-32d7ccbc7299 + JWT Auth token
//    - reservation marked as paid
// PATCH http://localhost:3000/api/reservations/84d37846-88f2-4692-b788-0c62e48b39db + JWT Auth token
//    - unauthorized access, since the user's id != reservation creator's id
// -- same can be done with DELETE to cancel the reservation
//
// GET http://localhost:3000/api/accommodations
//    - works for everyone with no authentication/authorisation needed



const app = express();
const router = express.Router();
const prisma = new PrismaClient();

// set Twig as our HTML renderer
app.set("view engine", "twig");

app.use(router);

// This section is optional and used to configure twig.
app.set("twig options", {
  allow_async: true, // Allow asynchronous compiling
  strict_variables: false,
});

// nesting the routers for accommodations
router.use("/accommodations", accommodationRouter);

// routing done via express router
router.use("/", (_req, _res, next) => {
  console.log(`[${Date.now()}] Incoming Request`);
  next();
});

router.get("/", (_req, res, _next) => res.redirect("../accommodations"));

router.get("/accommodations", async (_, res) => {
  try {
    const accommodationsData = await prisma.accommodation.findMany();
    res.render("index.twig", {
      accommodations: accommodationsData,
      topPickAccommodation: accommodationsData.length > 0 ? accommodationsData[0] : undefined
    });
  } catch (e) {
    res.render("index.twig", {
      accommodations: [],
      topPickAccommodation: undefined
    });
  }
});

const swaggerDocument = YAML.load(__dirname + '/../docs/swagger.yaml');

/**
 * Show Swagger UI for app Documentation
 */
app.use('/app-docs', swaggerUi.serve, swaggerUi.setup(swaggerDocument));

/**
 * Make express parse JSON in body and parse url encoded strings
 */
app.use(express.json());
app.use(express.urlencoded({ extended: true }));

/**
 * Serve static content from public directory for images
 */
app.use(express.static("public"));

app.use(cookieParser());

// after implementing JWT auth, basic auth checking using userId is only done at /login/token
// should be get/post/... instead of use? since it's not really middleware in this case, just using that endpoint
//   to get the JWT token - doesn't even use next(), since it always sends a response (but this way it covers all methods at least?)
app.use('/login/token', handleBasicAuth);

// we allow any users to use /api/accommodations for listing available accommodations
// temporarily disable to allow for easier functionality show-case
// app.use('/api', checkJwt.unless({path: '/api/accommodations'}));

app.use((err: any, _req: any, res: any, next: any) => {
  if (err.name === "UnauthorizedError") {
    res.status(401).send({
      status: "error",
      data: {},
      message: "Invalid token"
    });
  } else {
    next(err);
  }
});


/**
 * Resource accommodations
 */


// only the creator/host of the accommodation can update it
app.put('/api/accommodations/:id', accommodations.update);

// only logged in users (with valid JWT token) can create accommodations ..
// .. (and provided accommodation.userId in request body has to match)
app.post('/api/accommodations', accommodations.store);

// we allow anyone to display a list of accommodations
app.get('/api/accommodations', accommodations.list);


/**
 * Resource newsletter / subscriptions
 */
app.get('/api/newsletter/:id', newsletter.sendNewsletter)
app.post('/api/newsletter', newsletter.storeSubscription)

/**
 * Resource reservations
 */

// only the creator of the reservation can update it
app.patch('/api/reservations/:id', reservations.pay);

// only logged in users can create new reservations (and provided reservation.userId in request body has to match)
app.post('/api/reservations', reservations.store);

// only the creator of the reservation can cancel it
app.delete('/api/reservations/:id', reservations.cancel);

/**
 * Start listening on connections
 */
const port = process.env["PORT"] || 3000;
app.listen(
  port,
  () => console.log(`Express app is listening at http://localhost:${port}`)
)
