import { Request, Response } from 'express';
import { object, string, number, date, ValidationError } from 'yup';
import prisma from '../client';
import { sendAccommodationUpdatedMessage } from './newsletter';

/**
 * Request schema for full update / put
 */
const updateRequestSchema = object({
  name: string().required(),
  description: string().required(),
  mainPhoto: string().required(),  
  location: string().required(),
  addedAt: date().required(),
  userId: string().required().uuid(),
  price: number().positive().required()
});

const updateRequestSchemaSimple = object({
  name: string(),
  description: string(),
  mainPhoto: string(),
  location: string(),
  addedAt: date(),
  userId: string().uuid(),
  price: number().positive()
});

export const update = async (req: Request, res: Response) => {
  
  try {

    const accommodationId = req.params.id;

    const requestData = await updateRequestSchemaSimple.validate(req.body);
    const accommodation = await prisma.accommodation.findUnique({
      where: {
        id: accommodationId
      }
    });

    if (!accommodation) {
      return res.status(400).send({
        status: "error",
        data: {},
        message: "Accommodation with given ID not found"
      });
    }

    // if the user trying to update the accommodation isn't the host, deny the request
    // if (res.locals.jwtPayload.userId != accommodation.userId) {
    //   return res.status(403).send({
    //     status: "error",
    //     data: {},
    //     message: "Unauthorized access"
    //   })
    // }

    const accommodationUpdated = await prisma.accommodation.update({
      data: requestData,
      where: {
        id: accommodationId
      }
    })

    // send message to service bus - to send newsletter to users that are subscribed to this accommodation
    await sendAccommodationUpdatedMessage(accommodationId);

    return res.status(200).send({
      status: "success",
      data: accommodationUpdated,
      message: "Accommodation updated"
    })
  } catch (e) {
    if (e instanceof ValidationError) {
      return res.status(400).send({
        status: "error",
        data: e.errors,
        message: e.message
      });
    }

    return res.status(500).send({
      status: "error",
      data: {},
      message: "Something went wrong"
    });
  }
}

/**
 * Request schema
 */
 const storeRequestSchema = object({
  ...updateRequestSchema.fields,
  addedAt: date().default(new Date()),
});

/**
 * Store new accommodation
 */
export const store = async (req: Request, res: Response) => {

  try {
    const data = await storeRequestSchema.validate(req.body);

    // noone can create accommodation in someone else's name
    if (res.locals.jwtPayload.userId != data.userId) {
      // could also be 400 maybe? since the input data is invalid kinda?
      return res.status(403).send({
        status: "error",
        data: {},
        message: "Unauthorized access"
      })
    }

    const accommodation = await prisma.accommodation.create({
      data: data
    });

    return res.status(201).send({
      status: "success",
      data: accommodation,
      message: "Accommodation stored in system"
    })
  } catch (e) {
    if (e instanceof ValidationError) {
      return res.status(400).send({
        status: "error",
        data: e.errors,
        message: e.message
      });
    }

    return res.status(500).send({
      status: "error",
      data: {},
      message: "Something went wrong"
    });
  }
}

type Query = {
  price?: { lte?: string; gte?: string };
};

/**
 * Show all accommodations
 */ 
export const list = async (req: Request, res: Response) => {
  try {
    const query = req.query as Query;
    const queryLte = +(query.price?.lte ?? "") || Number.MAX_VALUE;
    const queryGte = +(query.price?.gte ?? "") || Number.MIN_VALUE;

    const accommodations = await prisma.accommodation.findMany({
      where: {
        AND: [
          {
            price: {
              lte: queryLte
            }
          },
          {
            price: {
              gte: queryGte
            }
          },
        ]
      }
    });

    return res.status(200).send({
      status: "success",
      data: accommodations,
      message: "List of all accommodations for given query parameters."
    })
  } catch (e) {
    return res.status(500).send({
      status: "error",
      data: {},
      message: "Something went wrong"
    });
  }
}
