import { ServiceBusClient } from "@azure/service-bus";
import { Request, Response } from 'express';
import prisma from '../client';
import { ValidationError, object, string } from "yup";


const AccommodationUpdated = 'AccommodationUpdated';
const UserSubscriptionsInfo = 'UserSubscriptionsInfo';

type AccommodationUpdatedType = typeof AccommodationUpdated;
type UserSubscriptionsInfoType = typeof UserSubscriptionsInfo;

type MessageType = AccommodationUpdatedType | UserSubscriptionsInfoType;

type Message = {
	type: MessageType;
	id: string;
}

type AccommodationUpdatedMessage = Message & {
	type: AccommodationUpdatedType;
}

type UserSubscriptionsInfoMessage = Message & {
	type: UserSubscriptionsInfoType
}

export const sendAccommodationUpdatedMessage = async (accommodationId: string) => {
	const message: AccommodationUpdatedMessage = { type: AccommodationUpdated, id: accommodationId };
	await sendMessage(message);
}

export const sendUserSubscriptionsInfoMessage = async (userId: string) => {
	const message: UserSubscriptionsInfoMessage = { type: UserSubscriptionsInfo, id: userId };
	await sendMessage(message);
}

// sends message to queue
const sendMessage = async (message: Message) => {
    const connectionString = process.env['CUSTOMCONNSTR_AZURE_SERVICE_BUS'] || "";
    const queueName = "newsletter-queue"
    const sbClient = new ServiceBusClient(connectionString);
	const sender = sbClient.createSender(queueName);

	try {

        let batch = await sender.createMessageBatch();
        batch.tryAddMessage({ body: JSON.stringify(message) });
        await sender.sendMessages(batch);

		console.log(`Sent a message to the queue: ${queueName}`);

		// Close the sender
		await sender.close();
	} finally {
		await sbClient.close();
	}

	return;
	
}

export const sendNewsletter = async (req: Request, res: Response) => {
    
    try {
        const userId = req.params.id;
        const user = await prisma.user.findUnique({
            where: {
            id: userId
            }
        });
    
        if (!user) {
            return res.status(400).send({
                status: "error",
                data: {},
                message: "User with given ID not found"
            });
        }

        await sendUserSubscriptionsInfoMessage(userId);

        return res.status(200).send({
            status: "success",
            data: {},
            message: "Newsletter sent"
        })
    } catch (e) {
        if (e instanceof ValidationError) {
        return res.status(400).send({
            status: "error",
            data: e.errors,
            message: e.message
        });
        }

        return res.status(500).send({
            status: "error",
            data: {},
            message: "Something went wrong"
        });
    }
}

const storeRequestSchema = object({
    userId: string().required().uuid(),
    accommodationId: string().required().uuid()
  });

// add new subscription
export const storeSubscription = async (req: Request, res: Response) => {
    
    try {
        const data = await storeRequestSchema.validate(req.body);
        const user = await prisma.user.findUnique({
            where: {
                id: data.userId
            }
        });
    
        if (!user) {
            return res.status(400).send({
                status: "error",
                data: {},
                message: "User with given ID not found"
            });
        }

        const accommodation = await prisma.accommodation.findUnique({
            where: {
                id: data.accommodationId
            }
        });
    
        if (!accommodation) {
            return res.status(400).send({
                status: "error",
                data: {},
                message: "Accommodation with given ID not found"
            });
        }

        const subscription = await prisma.subscription.findFirst({
            where: {
                userId: data.userId,
                accommodationId: data.accommodationId
            }
        });

        if (subscription != null) {
            return res.status(400).send({
                status: "error",
                data: {},
                message: "Subscription already exists"
            });
        }

        const newSubscription = await prisma.subscription.create({
            data: {
                userId: data.userId,
                accommodationId: data.accommodationId
            }
            // data: data
        });

        return res.status(201).send({
            status: "success",
            data: newSubscription,
            message: "Subscription stored in system"
        })
    } catch (e) {
        if (e instanceof ValidationError) {
        return res.status(400).send({
            status: "error",
            data: e.errors,
            message: e.message
        });
        }

        return res.status(500).send({
            status: "error",
            data: {},
            message: "Something went wrong"
        });
    }
}

export default sendMessage;