import { Request, Response } from 'express';
import { object, string, ValidationError, date, number } from 'yup';
import prisma from '../client';

export const pay = async (req: Request, res: Response) => {

  try {

    const reservation = await prisma.reservation.findUnique({
      where: {
        id: req.params.id
      }
    });

    if (!reservation) {
      return res.status(400).send({
        status: "error",
        data: {},
        message: "Reservation with given ID not found"
      });
    }

    // if the user trying to update the accommodation isn't the host, deny the request
    if (res.locals.jwtPayload.userId != reservation.userId) {
      return res.status(403).send({
        status: "error",
        data: {},
        message: "Unauthorized access"
      })
    }

    const reservationUpdated = await prisma.reservation.update({
      data: {
        isPaid: true,
      },
      where: {
        id: req.params.id
      }
    })

    return res.status(200).send({
      status: "success",
      data: reservationUpdated,
      message: "Reservation marked as paid"
    })
  } catch (e) {
    if (e instanceof ValidationError) {
      return res.status(400).send({
        status: "error",
        data: e.errors,
        message: e.message
      });
    }

    return res.status(500).send({
      status: "error",
      data: {},
      message: "Something went wrong"
    });
  }
}


const storeRequestSchema = object({
  from: date().required(),
  to: date().required(),
  guestCount: number().moreThan(1).required(),
  userId: string().required().uuid(),
  accommodationId: string().required().uuid(),
});

/**
 * Store new reservation
 */
export const store = async (req: Request, res: Response) => {

  try {
    const data = await storeRequestSchema.validate(req.body);

    if (data.from > data.to || data.from < new Date()) {
      return res.status(400).send({
        status: "error",
        data: {},
        message: "Reservations can only be made in the future and 'from' < 'to'"
      });
    }

    // check if there are no paid reservations for the chosen accommodation
    // in the time range [from ; to]
    const overlappingReservation = await prisma.reservation.findFirst({
      where: {
        AND: [
          {
            accommodationId: data.accommodationId,
          },
          {
            isPaid: true
          },
          {
            OR: [
              {
                // there is already a paid reservation that
                // starts in the middle of user's new reservation
                AND: [
                  { from: { gte: data.from } },
                  { from: { lte: data.to } }
                ]
              },
              {
                // there is already a paid reservation that
                // ends in the middle of user's new reservation
                AND: [
                  { to: { gte: data.from } },
                  { to: { lte: data.to } }
                ]
              }
            ],
          }
        ]
      }
    });

    // if there exists such a paid reservation that would overlap with user's, deny the request
    if (overlappingReservation) {
      return res.status(400).send({
        status: "error",
        data: {},
        message: "There already is a paid reservation for given accommodation in given time range"
      });
    }

    // noone can create reservation in someone else's name
    if (res.locals.jwtPayload.userId != data.userId) {
      // could also be 400 maybe? since the input data is invalid kinda?
      return res.status(403).send({
        status: "error",
        data: {},
        message: "Unauthorized access"
      })
    }
  
    // removed reservedAt/canceledAt/isPaid from the request, since these attrs shouldn't be set by the user
    const reservationData = {
      ...data,
      reservedAt: new Date(),
      canceledAt: null,
      isPaid: false
    };
    const reservation = await prisma.reservation.create({
      data: reservationData
    });

    return res.status(201).send({
      status: "success",
      data: reservation,
      message: "Reservation stored in system"
    })
  } catch (e: any) {
    if (e instanceof ValidationError) {
      return res.status(400).send({
        status: "error",
        data: e.errors,
        message: e.message
      });
    }

    return res.status(500).send({
      status: "error",
      data: {},
      message: "Something went wrong"
    });
  }
}


/**
 * Optional
 * 
 * @todo Cancel specific reservation, delete only sets deletedAt
 */
export const cancel = async (req: Request, res: Response) => {
  
  try {

    const reservation = await prisma.reservation.findUnique({
      where: {
        id: req.params.id
      }
    });

    if (!reservation) {
      return res.status(400).send({
        status: "error",
        data: {},
        message: "Reservation with given ID not found"
      });
    };

    // if the reservation is ongoing or old, we can't cancel it
    const reservationIsOld = reservation.to < new Date();
    const rservationIsOngoing = reservation.from <= new Date() && reservation.to > new Date();
    if (!reservation.canceledAt && (reservationIsOld || rservationIsOngoing)) {
      return res.status(400).send({
        status: "error",
        data: {},
        message: "Reservation can't be cancelled."
      });
    };

    const updatedReservation = await prisma.reservation.update({
      data: {
        canceledAt: new Date()
      },
      where: {
        id: req.params.id
      }
    });
  
    // if the user trying to cancel the reservation isn't the user who created the reservation, deny the request
    if (res.locals.jwtPayload.userId != reservation.userId) {
      return res.status(403).send({
        status: "error",
        data: {},
        message: "Unauthorized access"
      })
    }

    return res.status(200).send({
      status: "success",
      data: updatedReservation,
      message: "Reservation cancelled"
    })
  } catch (error) {
    return res.status(500).send({
      status: "error",
      data: {},
      message: "Something went wrong"
    });
  }
}